#!/usr/bin/python

"""
hsplit.py - Go through the netcdf footprint files and create the H slices

Do this by stepping through the observations.  For each obs, read in the
netcdf file.  Then step through the H block timesteps, find the footprint values
that match the timestep, and write any nonzero values to a temporary text H slice file.
Each subsequent observation footprint will append to the H text files.
After stepping through all timesteps, reformat the text files to binary format.

---
for each obs:
	read footprint
	for each timestep
		append values to text files, H1, H2, H3, ...

reformat files to binary format
---

This way we read each netcdf file only 1 time.

Requires:
	receptor list - list of footprint file names to use
	footprint files - stilt footprint files
	landmask - landmask file which defines land cells in the domain of interest.  See mklm.py
	config file - Configuration file

Output:
	H strip binary files.  There is one file for each timestep.  Files are in binary format.
"""


import os
import sys
import datetime
import netCDF4
import shutil
import getopt

from utils import *


####################################################
def usage():

        print """"
Usage: hsplit.py [-c configfile] receptor_file

where 
        receptor_file is file with list of footprint file names.
	-c configfile : configuration file to use.  Default is 'config.ini'.
	"""

####################################################
def get_footprint(filename):
        """
        Return from the netcdf file 'filename' the footprint grid.
	This the the netcdf variable 'foot' from the stilt netcdf footprint files.
        """

        try:
                ds = netCDF4.Dataset(filename)
        except:
                print >> sys.stderr, "Error trying to read netcdf file %s." % filename
                return None

        g = ds.variables['foot1']
        grid = g[:]
	lats = ds.variables['foot1lat'][:]
	lons = ds.variables['foot1lon'][:]
	dy = lats[1] - lats[0]
	dx = lons[1] - lons[0]

	s = list(g.dimensions)
	n1 = s.index("foot1lat")
	n2 = s.index("foot1lon")

        f1dates = ds.variables['foot1date'][:]

        ds.close()

	# Convert the foot1date array from days since jan 1, 2000 to actual datetime
	dates = []
	refdate = datetime.datetime(2000, 1, 1)
	for d in f1dates:
		date = refdate + datetime.timedelta(days=d)
		dates.append(date)

	# if 'foot1lon' dimension comes before 'foot1lat', swap the dimensions
	# some netcdf files store the footprint as (nstep x lon x lat).
	# We need (nstep x lat x lon).  Switch lat, lon around here
	if n2 < n1:
                b = numpy.empty( (grid.shape[0], grid.shape[2], grid.shape[1]) )
                nrows = grid.shape[0]
                for i in range(nrows):
                        a = grid[i]
                        b[i] = a.T

                grid = b

        # reduce grid to 1x1 degree (assumes dx = dy. would need to change if not)
        if dx == 0.5:
                c =  grid[:, 0:-1:2, 0:-1:2] + grid[:, 0:-1:2, 1::2] + grid[:, 1::2, 0:-1:2] + grid[:, 1::2, 1::2]
	else:
		c = grid

        return c, dates

####################################################

write_binary = True
configfile = "config.ini"

try:
        opts, args = getopt.getopt(sys.argv[1:], "c:")
except getopt.GetoptError, err:
        # print help information and exit:
        print str(err) # will print something like "option -a not recognized"
        usage()
        sys.exit()

for o, a in opts:
        if o == "-c":
                configfile = a
#	elif o == "-b":
#		write_binary = True

if len(args) != 1:
	usage()
	sys.exit()

listfile = args[0]


config = getConfig(configfile)
t0 = datetime.datetime.now()


#!!!! new landmask created using mklm.py
landmaparr = numpy.load(config["landmask_file"])

# Loop through every receptor file
try:
	f = open(listfile)
except:
	sys.exit("Cannot open receptor file %s." % listfile)

# Temporary directory to hold text files 
tmpdir = config["h_tmp_dir"]
if os.path.exists(tmpdir):
	# remove any files in tmpdir
	shutil.rmtree(tmpdir)

# create tmpdir
os.makedirs(tmpdir)

obsnum = 0
for line in f:

	filename = line.strip()

	t1 = datetime.datetime.now()

	# read the netcdf file for the obs.
	# grid returned is (ntimesteps x lat x lon) for 1 degree grid
	grid, griddates = get_footprint(filename)

	# extract land cells.  This also changes shape of grid to 2d numobs x numlandcells
	grid = grid[:, landmaparr[0], landmaparr[1]]

	# Loop through each of the grid dates, find ones with dates in desired range.
	# We're going backwards in time, so skip dates after the end date, quit when before start date.
	# If found, extract non zero grid indices, write those to file for correct inversion timestep
	# If inversion timestep > footprint timestep, then sum up the footprint values that fit in
	# the inversion timestep window
	# e.g. if inversion timestep is 3 hours, then sum up hours 0, 1, 2 for timestep 1; 3, 4, 5 for timestep 2 ...
	xx = {}
	for gidx, gd in enumerate(griddates):
		if gd < config["sdate"]: break
		if gd >= config["edate"]: continue

		td = gd - config["sdate"]			# time since inversion start date,
		nhours = td.days*24 + td.seconds/3600		# in number of hours
		ntimestep = nhours / config["hrsperstep"] + 1  	# inversion timestep number
		nz = grid[gidx].nonzero()[0]			# location of non zero data points
		for cellnum in nz:
			val = grid[gidx, cellnum]
			if val > -3e34:
				if (ntimestep,cellnum) not in xx: xx[(ntimestep, cellnum)] = 0  # create a dict entry for this timestep and cell number
				xx[(ntimestep,cellnum)] += val   # add value to this timestep,cellnum


	# make a dict that is indexed by timestep only, 
	# i.e. contains all data for each timestep for this observation 
	b = {}
	for timestep, cellnum in xx.keys():
		if timestep not in b: b[timestep] = []
		b[timestep].append((obsnum, cellnum, xx[(timestep,cellnum)]))


	# write data to temporary text files, 
	for ntimestep in b.keys():
		file = "%s/H%04d.txt" % (tmpdir, ntimestep)
		f = open(file, "a")
		for (obsnum, cellnum, val) in b[ntimestep]:
			f.write("%d %d %15.8e\n" % (obsnum, cellnum, val) )
		f.close()


	t2 = datetime.datetime.now()
	print "Finished obs num ", obsnum,  t2-t1, t2-t0
	obsnum += 1

f.close()


num_obs = obsnum

# now reformat the temporary text files into final binary files

# set the directory where the H slices will be stored.
hdir = config["workdir"] + "/" + config["hdir"] 
if os.path.exists(hdir):
	# remove any files in the hdir.
	shutil.rmtree(hdir)

# create the h directory
try:
	os.makedirs(hdir)
except:
	sys.exit("Can't make directory for h files: %s" % hdir)

for i in range(1,config["ntimesteps"]+1):

	file = "%s/H%04d.txt" % (tmpdir, i)	# text file to read from
	filename = "%s/H%04d.bin" % (hdir, i)	# binary file to write to
	print "convert ", file, "to", filename

	if os.path.exists(file):

		h = numpy.loadtxt(file, dtype=[('obs', numpy.int32), ('cell', numpy.int32), ('val', float)])

		makebinsparse(filename, h, num_obs, config["nlandcells"])

t2 = datetime.datetime.now()
print "Finished. Total elapsed time", t2-t0
