#!/usr/bin/python

"""
bayesian inversion

 Solve the equation
  shat = sprior + HQT*inverse(HQHT+R)*(z-hsp)

Input files:
	HQHT.npy, HQHT_pbl.npy, HQHT_trans.npy, HQHT_ftrop.npy
	r.txt
	zhsp.txt
	sprior.npy
	All HQ slice files

Output files:
	hqhti.npy
	hqht_zhsp.npy
	qhts.npy
	hshat.txt
	shat_flux.npy, shat_flux.nc
	shat_pbl.npy, shat_pbl.nc
	shat_trans.npy, shat_trans.nc
	shat_ftrop.npy, shat_ftrop.nc

"""


import sys
import numpy
import getopt

from utils import *


#########################################################################
def usage():

        print """
inversion.py [-c configfile] [-b] spriorfile rfile

where 

   Required arguments
	spriorfile : file name with sprior data
	rfile : file name with R data

   Optional arguments
	-c : Specify the configuration file to use. Default is 'config.ini'.
	-b : If set, use boundary value optimization files.

        """

####################################################################
def qht_s(config, hqht_zhsp):
	""" compute hqt*[inverse(hqht+r)*(z-h*sp)]

	inverse(hqht+r)*(z-h*sp) is given as the array hqht_zhsp.
	hq is read in from files created previously.
	"""

	global num_boundary_cells

	ntimesteps = config["ntimesteps"]

	#--------------
	if include_boundary:
		# the 4200 is the number of cells in the bounday optimization grid * 3 levels (pbl, trans, ftrop)
		qhts = numpy.zeros( (ntimesteps, config["locations"] + num_boundary_cells*3) )
		hqsum = numpy.zeros( (config["num_obs"], config["locations"] + num_boundary_cells*3) )
	else:
		qhts = numpy.zeros( (ntimesteps, config["locations"]) )
		hqsum = numpy.zeros( (config["num_obs"], config["locations"]) )
	#--------------

	t0 = datetime.datetime.now()
	for i in range(ntimesteps):
		t1 = datetime.datetime.now()

		# hq slice files are numbered 1 .. ntimesteps
		# hq shape is #obs x #cells
		hqfile = config["workdir"] + "/" + config["hqdir"] + "/HQ%04d.npy" % (i+1)
		hq = numpy.load(hqfile)

		#--------------
		if include_boundary:
			# convert 0-735 3 hour timestep to 1-92 boundary daily timestep number.
			bn = i/8 + 1
			hqpblfile = "HQpbl/HQ%04d.npy" % bn
			hqtransfile = "HQtrans/HQ%04d.npy" % bn
			hqftropfile = "HQftrop/HQ%04d.npy" % bn
			hqpbl = numpy.load(hqpblfile)
			hqtrans = numpy.load(hqtransfile)
			hqftrop = numpy.load(hqftropfile)
			hq = numpy.hstack( (hq, hqpbl, hqtrans, hqftrop) )
		#--------------

		qhts[i] = numpy.dot(hq.T,hqht_zhsp)

		t2a = datetime.datetime.now()
		print "Step %4d" % (i+1), "of ", ntimesteps, "done. Step time ", t2a-t1, "Total elapsed time ", t2a-t0

	return qhts


####################################################################


if __name__ == '__main__':


	numpy.set_printoptions(precision=6, edgeitems=10, linewidth=150)

	configfile = "config.ini"
	include_boundary = False
	resume = False

	try:
		opts, args = getopt.getopt(sys.argv[1:], "bc:", ["resume"])
	except getopt.GetoptError, err:
		# print help information and exit:
		print str(err) # will print something like "option -a not recognized"
		usage()
		sys.exit()

	for o, a in opts:
		if o == "-b":
			include_boundary = True
		elif o == "-c":
			configfile = a
		elif o == "--resume":
			resume = True

	print args
	if len(args) != 2:
		usage()
		sys.exit()

	spriorfile = args[0]
	rfile = args[1]

	config = getConfig(configfile)

	tstart = datetime.datetime.now()


	# Get hqht calculated previously.
	# shape of hqht is numobs x numobs
	print "Getting hqht..."
	filename = config["workdir"] + "/" + config["hqdir"] + "/HQHT.npy"
	hqflux = numpy.load(filename)
	numobs = hqflux.shape[0]
	config["num_obs"] = numobs

	####################
	if include_boundary:
		# combine the 4 different hqht's (flux, pbl, trans, free troposphere)
		config_pbl = getConfig("config_pbl.ini")
		config_trans = getConfig("config_trans.ini")
		config_ftrop = getConfig("config_ftrop.ini")

		num_boundary_cells = config_pbl["ncells"]

		filename = config_pbl["workdir"] + "/" + config_pbl["hqdir"] + "/HQHT.npy"
		print filename
		hqpbl = numpy.load(filename)

		filename = config_trans["workdir"] + "/" + config_trans["hqdir"] + "/HQHT.npy"
		print filename
		hqtrans = numpy.load(filename)

		filename = config_ftrop["workdir"] + "/" + config_ftrop["hqdir"] + "/HQHT.npy"
		print filename
		hqftrop = numpy.load(filename)

		hqht = hqflux + hqpbl + hqtrans + hqftrop
	else:
		hqht = hqflux
	####################


	print "Adding r to hqht..."
#        rfile = config["workdir"] + "/" + config["r_file"]
        r = numpy.loadtxt(rfile)
	for i in range(numobs):
		hqht[i,i] = hqht[i,i] + r[i]


	# calculate inverse of hqht
	# hqhti is needed later in aposterior variance calculations
	if resume:
		print "Loading hqhti..."
		hqhti = numpy.load("hqhti.npy")
	else:
		print "Calculate inverse of hqht..."
		hqhti = numpy.linalg.inv(hqht)
		numpy.save("hqhti", hqhti)


	print "Getting zhsp..."
	zfile = config["workdir"] + "/" + config["z_file"]
	zhsp = numpy.loadtxt(zfile)

	if resume:
		print "Loading hqht_zhs..."
		hqht_zhsp = numpy.load("hqht_zhsp.npy")
	else:
		print "Calculating hqhti*zhsp..."
		hqht_zhsp = numpy.dot(hqhti,zhsp)
		numpy.save("hqht_zhsp", hqht_zhsp)
	
	# compute qht*[inverse(hqht+r)*(z-h*sp)]
	print "Calculating qhts..."
	qhts = qht_s(config, hqht_zhsp) 
	print "qhts shape is ", qhts.shape
	numpy.save("qhts", qhts)

	print "Getting sprior..."
	try:
		sp = numpy.load(spriorfile)
	except:
		try:
			landmask = numpy.load(config["landmask_file"])
			sp = getSpriorNetcdf(spriorfile, "sprior", landmask)
		except:
			print "Can not read sprior file."
			sys.exit()


	if include_boundary:
		a = numpy.zeros( (config["ntimesteps"], num_boundary_cells*3) )
		sp = numpy.hstack((sp, a))
	print "sp shape is ", sp.shape


	print "Calculate shat..."
	shat = sp + qhts

	print "Saving shat in netcdf format..."
        if include_boundary:
		numpy.save("shat_boundary", shat)

		shat_flux = shat[:, 0:config["nlandcells"]]
		shat_pbl  = shat[:, config["nlandcells"]:config["nlandcells"] + num_boundary_cells]
		shat_trans  = shat[:, config["nlandcells"] + num_boundary_cells:config["nlandcells"] + 2*num_boundary_cells]
		shat_ftrop  = shat[:, config["nlandcells"] + 2*num_boundary_cells:config["nlandcells"] + 3*num_boundary_cells]
		numpy.save("shat_flux", shat_flux)
		numpy.save("shat_pbl", shat_pbl)
		numpy.save("shat_trans", shat_trans)
		numpy.save("shat_ftrop", shat_ftrop)
		make_ncdf("shat_flux.nc", "shat", shat_flux, config)
		make_ncdf("shat_pbl.nc", "shat", shat_pbl, config_pbl)
		make_ncdf("shat_trans.nc", "shat", shat_trans, config_trans)
		make_ncdf("shat_ftrop.nc", "shat", shat_ftrop, config_ftrop)

        else:
		numpy.save("shat", shat)
		make_ncdf("shat.nc", "shat", shat, config)


	print "Making hshat..."
	if include_boundary:
		hshat = convolve(shat_flux, config)
		hshat_pbl = convolve(shat_pbl, config_pbl)
		hshat_ftrop = convolve(shat_ftrop, config_ftrop)
		hshat_trans = convolve(shat_trans, config_trans)
		numpy.savetxt("hshat_pbl.txt", hshat_pbl)
		numpy.savetxt("hshat_trans.txt", hshat_trans)
		numpy.savetxt("hshat_ftrop.txt", hshat_ftrop)
		numpy.savetxt("hshat_flux.txt", hshat)
	else:
		hshat = convolve(shat, config)
		numpy.savetxt("hshat.txt", hshat)


	tdone = datetime.datetime.now()
	print "Finished. Elapsed time", tdone-tstart

